/* Sample MPI code for parallelized JDBSYM.
 *
 * To build:
 *	mpicc -O2 -Wall -c sample-mpi.c
 *	mpif77 -o sample-mpi sample-mpi.o -ljdbsym -llapack -lblas -lm
 */

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

#include <mpi.h>
#include "jdbsym.h"

static int nglobal = 100;
static int n, myid, nprocs, prev, next;

/* function computing y := A x */
static void
matveca(double *x, double *y)
{
  int i;
  double a = 0, b = 0;
  MPI_Status status;

  assert(n >= 2);

  MPI_Sendrecv(x, 1, MPI_DOUBLE, prev, 111, &a, 1, MPI_DOUBLE, prev, 111,
		MPI_COMM_WORLD, &status);
  MPI_Sendrecv(x+n-1, 1, MPI_DOUBLE, next, 111, &b, 1, MPI_DOUBLE, next, 111,
		MPI_COMM_WORLD, &status);

  *(y++) = -a + 2. * x[0] - x[1];
  for (i = 1; i < n - 1; i++, x++)
	*(y++) = -x[0] + 2. * x[1] - x[2];
  *(y++) = -x[0] + 2. * x[1] - b;
}

/* function computing y := B x */
static void
matvecb(double *x, double *y)
{
  int i;
  double d = 1.0 / ((nglobal + 1) * (nglobal + 1));

  for (i = 0; i < n; i++)
	*(y++) = d * *(x++);
}

static void
gsum(double *sbuf, double *rbuf, int count, void *user_data)
{
    MPI_Allreduce(sbuf, rbuf, count, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
}

int
main(int argc, char **argv)
{
  double *Q, *lambda;
  int k_conv, it;

  /* JD parameters */
  double tau = 0.0;
  double eps = 1e-10;
  double eps_tr = 1e-4;
  double toldecay = 2.0;

  int blksize = 1;
  int blkwise = 0;
  int kmax = 5;
  int jmax = 2*kmax;
  int jmin = kmax;
  int itmax = 150;
  int mvmax = 100;
  int clvl = 1;
  int linsolver = LINSOLV_CGS;
  int optype = OP_UNSYM1;
  int strategy = 0;
  int i;

  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &nprocs);
  MPI_Comm_rank(MPI_COMM_WORLD, &myid);

  /* compute local vector size */
  n = nglobal / nprocs + (myid < nglobal % nprocs ? 1 : 0);
  jdbsym_set_parms(NULL, gsum, nglobal, nprocs, myid);
  prev = (myid == 0 ? MPI_PROC_NULL : myid - 1);
  next = (myid == nprocs - 1 ? MPI_PROC_NULL : myid + 1);

  Q = (double *)malloc(n * kmax * sizeof(double));
  lambda = (double *)malloc(kmax * sizeof(double));
  
  jdbsym(n, tau, eps, kmax, jmax, jmin, itmax, blksize, blkwise,
	 0, NULL, linsolver, optype, mvmax, eps_tr, toldecay, strategy,
	 clvl, &k_conv, Q, lambda, &it, matveca, matvecb, NULL);

  n = nglobal + 1;
  printf("converged eigenvalues: %d\n", k_conv);
  for (i = 0; i < k_conv; i++) {
    tau = 2.*(1.-cos((i + 1) * M_PI/n)) * n * n;
    printf("error[%d] = %lg\n", i + 1, lambda[i] - tau);
  }

  MPI_Finalize();

  return 0;
}
